
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
import pandas as pd
import numpy as np


def adjRsquare(pre_y,r,vars):
  l=len(pre_y)
  return 1 - (1-r2_score(pre_y, r)) * (l-1)/(l-len(vars)-1)
def getVIFs(vars,df):
  # result={}
  # for i in vars:
  #   pre_y=df[i].to_numpy()
  #   df_x=pd.concat([df[[j for j in vars if j!=i]],
                  
  #           ],axis=1)
  #   pre_X=df_x.to_numpy()
  #   # weight=df[weight_col].to_numpy()

  #   #hago la regresion
  #   reg = LinearRegression().fit(pre_X, pre_y)   #############OJOTA ,weight
  #   #reg = LinearRegression().fit(pre_X, pre_y)
  #   result[i]=1/(1-r2_score(pre_y, reg.predict(pre_X)) )
  #   result[i]=1/(1-adjRsquare(pre_y, reg.predict(pre_X),vars) )
  
  duff=df[vars]
  df_cor = duff.corr()
  vif_matrix=pd.DataFrame(np.linalg.inv(duff.corr().values), index = df_cor.index, columns=df_cor.columns)
  vif_diag=pd.Series(np.diag(vif_matrix), index=vif_matrix.columns)
  
  return vif_diag.to_dict()

#viffs
def getIndexOfCol3(var,allVIFs):
  counter=0
  index=[]
  for j in allVIFs: 
    if var in j.keys():
      index.append(counter)
    counter+=1
  return index
def getFullCombination(boolArray,colls):
  counter=0
  final=[]
  for j in boolArray:
    if j==1:
      final.append(colls[counter])
    counter+=1
  return final


#meta reg
def getIndexOfCol(var,allBetas):

  counter=0
  index=[]
  for j in allBetas:
    if var in j.keys():
      index.append(counter)
    counter+=1
  return index



from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
import numpy as np
from tqdm import tqdm
import pandas as pd
import plotly.express as px
from .low_level import getFullCombination
from IPython.display import display,HTML


def importance_test(df1,y_col,weight_col=None,control_cols='not',const_cols='not'):
  df=df1.copy()
  if weight_col:
    df=df.loc[df[weight_col]>0]


  if const_cols=='not':
    const_cols=[i for i in df.columns if ('.hol' in i.lower()) or ('.mkt' in i.lower())]
  if control_cols=='not':
    control_cols=[i for i in df.columns if i not in const_cols+[y_col,weight_col,'Geographies','Weeks','non_lin_pred','lin_pred']]
  #
  weight=df[weight_col].to_numpy() if weight_col else None

  
  x_cols=control_cols+const_cols
  X_rand=np.ones([len(control_cols),len(control_cols)])-np.eye(len(control_cols))
  allBetas=[]
  y_miniReg=df[y_col].to_numpy()
  #meta
  allLosses=[]
  for i in tqdm(X_rand):
    # print(i)
    randomX_cols=getFullCombination(i,control_cols)
    df_x=pd.concat([df[randomX_cols],df[const_cols]

          ],axis=1)
    X_miniReg=df_x.to_numpy()

    #hago la regresion
    reg = LinearRegression().fit(X_miniReg, y_miniReg, weight)
    betas={df_x.columns[i]:reg.coef_[i] for i in range(len(df_x.columns))}
    allLosses.append(np.mean((reg.predict(X_miniReg)-y_miniReg)**2))
    allBetas.append(betas)
  #get base
  pre_y=df[y_col].to_numpy()
  df_x=pd.concat([df[x_cols],
          ],axis=1)
  pre_X=df_x.to_numpy()
  
  reg = LinearRegression().fit(pre_X, pre_y,weight)
  topred=pd.concat([df[x_cols],
          ],axis=1)
  topred_np=topred.to_numpy()
  df['lin_pred']=reg.predict(topred_np)
  baseError=np.mean((df['lin_pred'].to_numpy()-df[y_col].to_numpy())**2)
  ress=pd.DataFrame(dict(cols=control_cols, imp=np.array(allLosses)-baseError))
  ress.sort_values('imp').plot('cols', 'imp', 'barh')
  


  x_cols=control_cols+const_cols
  pre_y=df[y_col].to_numpy()
  df_x=pd.concat([df[x_cols],
          ],axis=1)
  pre_X=df_x.to_numpy()

  rf = RandomForestRegressor(1000, min_samples_leaf=5,n_jobs=-1)
  rf.fit(pre_X, pre_y);
  
  linn = LinearRegression().fit(pre_X, pre_y,weight)

  topred=pd.concat([df[x_cols],],axis=1)
  topred_np=topred.to_numpy()
  df['non_lin_pred']=rf.predict(topred_np)
  df['lin_pred']=linn.predict(topred_np)

  fulltotal=df.groupby(['Weeks']).mean()
  fulltotal=fulltotal.reset_index()
  fig=px.line(fulltotal,x='Weeks',y=[y_col,'non_lin_pred','lin_pred'])
  fig.show()
  #guardar_fig(fig,"bg overfitteado con vars de covid")
  display(HTML('<h1>First Linear, Second Non-Linear<h1>'))
  np.mean((df['non_lin_pred'].to_numpy()-df[y_col].to_numpy())**2)
  ress=pd.DataFrame(dict(cols=x_cols, imp=rf.feature_importances_))
  # ress.loc[ress["imp"]>0].plot('cols', 'imp', 'barh')
  ress.sort_values('imp').loc[ress["imp"]>0.001].plot('cols', 'imp', 'barh')


import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from tqdm import tqdm
from .low_level import getFullCombination,getIndexOfCol
import plotly.express as px
from IPython.display import display

class Meta_Reg:
    def __init__(self, df1,y_col,stochQ=10000,weight_col=None,control_cols=None,const_cols=None,solve_problems=False):
        df = df1.copy()
        if weight_col:
            df=df.loc[df[weight_col]>0]
        if not const_cols:
            const_cols=[i for i in df.columns if ('.hol' in i.lower()) or ('.mkt' in i.lower())]
        if not control_cols:
            control_cols=[i for i in df.columns if i not in const_cols+[y_col,weight_col,'Geographies','Weeks']]
        weight=df[weight_col].to_numpy() if weight_col else None

        #pre meta reg
        x_cols=control_cols+const_cols
        X_rand=np.random.randint(2, size=[stochQ,len(control_cols)])
        allBetas=[]
        y_miniReg=df[y_col].to_numpy()
        allLosses=[]

        for i in tqdm(X_rand):
            # print(i)
            randomX_cols=getFullCombination(i,control_cols)
            # print(x_col)
            # print(getFullCombination(i,x_col))
            df_x=pd.concat([df[randomX_cols],df[const_cols]

                    ],axis=1)
            X_miniReg=df_x.to_numpy()

            #hago la regresion
            reg = LinearRegression().fit(X_miniReg, y_miniReg, weight)
            betas={df_x.columns[i]:reg.coef_[i] for i in range(len(df_x.columns))}
            allLosses.append(np.mean((reg.predict(X_miniReg)-y_miniReg)**2))

            allBetas.append(betas)
            # print(betas)
        if solve_problems:
        #problems with reg
            px.histogram(np.log(allLosses)).show()
            px.histogram(np.log(allLosses))
            indices=[i for i in range(len(allLosses)) if np.log(allLosses[i])<0]
            allBetas=list(np.array(allBetas)[indices])
            allLosses=list(np.array(allLosses)[indices])
            X_rand=X_rand[indices]
            px.histogram(np.log(allLosses)).show()


        df['periods']=df['Weeks'].dt.year
        selectedPeriod=max(df['periods'].unique())
        supp_df=df.groupby(['periods']).mean()
        medias=supp_df.loc[selectedPeriod].to_dict()
        #to save
        self.allLosses=allLosses
        self.allBetas=allBetas
        self.df=df
        self.X_rand=X_rand
        self.stochQ=stochQ
        self.control_cols=control_cols
        self.const_cols=const_cols
        self.x_cols=x_cols
        self.y_miniReg=y_miniReg
        self.medias=medias

        display(control_cols)
    
    def beta_reg(self,var,control_var=[]):
        allBetas=self.allBetas
        X_rand=self.X_rand
        allLosses=self.allLosses
        medias=self.medias
        control_cols=self.control_cols
        control_var=control_var+[var]
        print(var)
        toHist=[]
        fast=set.intersection(*map(set,[getIndexOfCol(i,allBetas) for i in control_var]))
        for i in getIndexOfCol(var,allBetas):
            if i in fast:
                toHist.append({"value":allBetas[i][var],"control":"controled","error":allLosses[i],'contribution':allBetas[i][var]*medias[var]*100})
            else:
                toHist.append({"value":allBetas[i][var],"control":"non_controled","error":allLosses[i],'contribution':allBetas[i][var]*medias[var]*100})
        tograph=pd.DataFrame(toHist)

        #px.histogram(tograph,x="value",color="control",nbins=1000).show()

        metaY=np.array([allBetas[i][var] for i in getIndexOfCol(var,allBetas)])
        metaX=X_rand[getIndexOfCol(var,allBetas)]


        #px.histogram(x=metaY,nbins=1000).show()



        metaReg = LinearRegression(fit_intercept=False).fit(metaX, metaY)




        meta_betas={control_cols[i]:metaReg.coef_[i] for i in range(len(control_cols))}
        #px.scatter(x=[np.array(allLosses)[i] for i in getIndexOfCol(var)],y=metaY).show()
        #px.scatter(tograph,x="error",y="value",color="control").show()
        px.histogram(tograph,x="contribution",color="control",nbins=1000).show()
        px.scatter(tograph,x="error",y="contribution",color="control").show()
        meta_betas
        #print(meta_betas[var])
        #{k: v for k, v in sorted(meta_betas.items(), key=lambda item: item[1])}
        display({k: v*medias[var]*100 for k, v in sorted(meta_betas.items(), key=lambda item: item[1])})
        #getT(metaReg,df_x,metaX,metaY)


import pandas as pd
from scipy.cluster.hierarchy import linkage, dendrogram, fcluster
from scipy.spatial.distance import squareform
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import plotly.express as px
from tqdm import tqdm
from IPython.display import display


from .low_level import getFullCombination, getVIFs, getIndexOfCol3



#############################################################################################################################################################
#second page


def vif_distance(df,tounderstand,stochQ=1000):   #transform meta vif analisys to block modeling
  #'searchcont
  X_rand=np.random.randint(2, size=[stochQ,len(tounderstand)])
  X_rand=X_rand[np.where(X_rand.sum(axis=1)>=2)]
  allVIFs=[]

  for i in tqdm(X_rand):
    # print(i)
    randomX_cols=getFullCombination(i,tounderstand)
    allVIFs.append(getVIFs(randomX_cols,df))

  for i in allVIFs:
    for j in i.keys():
      if i[j]>3000:
        i[j]=3000
  
  interest=tounderstand

  impacten=tounderstand

  fullTest=tounderstand
  matoPato=[]
  for k in fullTest:
    var=k#x_cols[ind]
    metaY=np.array([allVIFs[i][var] for i in getIndexOfCol3(var,allVIFs)])
    metaX=X_rand[getIndexOfCol3(var,allVIFs)]
    metaReg = LinearRegression(fit_intercept=False).fit(metaX, metaY)
    # meta_betas
    normed_coef=(metaReg.coef_-metaReg.coef_.mean())/metaReg.coef_.std()
    normed_coef=metaReg.coef_
    matoPato.append(list(normed_coef))
  duff=pd.DataFrame(np.array(matoPato).transpose())
  duff.columns=fullTest
  duff.index=impacten

  duff.values[[np.arange(duff.shape[0])]*2] = 0

  to_test=interest

  to_test_index=impacten

  filtered_duff=duff.loc[to_test_index,to_test]
  simil=np.minimum(filtered_duff.to_numpy(),filtered_duff.to_numpy().T)

  simil[simil < 0] = 0
  disimil=1/simil
  disimil[disimil > 200] = 200
  disimil=np.log(disimil+1)
  n = disimil.shape[0]
  disimil[range(n), range(n)] = 0
  disimil=disimil/np.max(disimil)

  simil=1-disimil
  similitude=pd.DataFrame(simil)
  similitude.columns =tounderstand
  similitude.index=tounderstand
  return similitude



def collinearity_test(df1,tounderstand='not',distance="corr"):   #full test of collinearity
  df=df1.copy()
  if tounderstand=='not':
    const_cols=[i for i in df.columns if ('.hol' in i.lower()) or ('.mkt' in i.lower())]
    tounderstand=[i for i in df.columns if i not in const_cols+['Geographies','Weeks','non_lin_pred','lin_pred']]
  to_test=tounderstand

  display(getVIFs(tounderstand,df))

  # set figure size
  correlation=df[to_test].corr().fillna(0)
  filter=0
  corr_numpy=correlation.to_numpy()
  indices=corr_numpy.argsort()[:,-2]
  corr_numpy
  finalIndices=[]
  for i in range(len(corr_numpy)):
    value=corr_numpy[i,indices[i]]
    if value>filter or value<-filter:
      finalIndices.append(i)
  finalIndices
  finalCorr=correlation.iloc[finalIndices,finalIndices]

  plt.figure(figsize=(10,7))

  # Generate a mask to onlyshow the bottom triangle
  mask = np.triu(np.ones_like(finalCorr, dtype=bool))

  # generate heatmap
  sns.heatmap(finalCorr, annot=True, mask=mask, vmin=-1, vmax=1,cmap="rocket_r")
  plt.title('Correlation Coefficient Of Predictors')
  plt.show()


  
  data=df[to_test]

  #block distance
  #correlations
  correlations=df[to_test].corr().fillna(0)
  if distance=="corr":
    
    similarity=abs(correlations)
  elif distance=="vif":
    similarity=vif_distance(df,tounderstand,stochQ=1000)
  

  plt.figure(figsize=(12,8))
  dissimilarity = 1 - similarity
  Z = linkage(squareform(dissimilarity), 'complete')

  dendrogram(Z, labels=to_test, orientation='top', 
            leaf_rotation=90);
  plt.tight_layout()
  plt.savefig('maybe.svg', format='svg', dpi=1200)

  ########third
  px.line(((df.groupby('Weeks').mean()-df.groupby('Weeks').mean().mean())/df.groupby('Weeks').mean().std()).reset_index(),x='Weeks',y=tounderstand).show()





import pyspssio
import random
import clevercsv
import pyreadstat
import pandas as pd
from google.colab import drive
from IPython.display import display

def input_path(message=""):  #translate copy path of windows
  raw_s=input(message)
  raw_s=raw_s.replace(r'"','')
  print("/content/drive/Shareddrives/"+'/'.join(raw_s.split("\\")[2:]))
  return "/content/drive/Shareddrives/"+'/'.join(raw_s.split("\\")[2:])
   
def get_data(path=False,sheet_name=0,usecols=None):  #easy get data from MES csv   ###############################################  path y sheet_name
    #import drive
    drive.mount("/content/drive")
    #transform str and get doc
    if path:
      newpath=path
    else:
      newpath=input_path('Input path:')
    
    if (newpath.split("/")[-1].split(".")[-1] == "xlsx") or (newpath.split("/")[-1].split(".")[-1] == "xls"):
        df=pd.read_excel(newpath,sheet_name=sheet_name)
    elif (newpath.split("/")[-1].split(".")[-1] == "csv"):
        df=pd.read_csv(newpath)
    elif (newpath.split("/")[-1].split(".")[-1] == "sav"):
        df=pd.read_spss(newpath,usecols=usecols)
        # try:
        #    df=df.rename(columns={"WEEK": "Weeks", "MKT": "Geographies"})
           
        # except:
        #    pass
    else:
        raise ValueError('Format Not Found')
    try:
      df=df.drop("Unnamed: 0",axis=1)
    except:
       pass
    try:
      df["Weeks"]=pd.to_datetime(df["Weeks"])
    except:
      pass
    
    return df
#"G:\Shared drives\Team members\Agustin\merk\00Q3\1websls\data.csv"
###########################################################################################################################################################


def to_sav(option='default',redundancy=False):  #try to read csv and parse it to sav
    #import drive
    drive.mount("/content/drive")
    newpath=input_path('Input path:')
    if (newpath.split("/")[-1].split(".")[-1] == "xlsx") or (newpath.split("/")[-1].split(".")[-1] == "xls"):
        df=pd.read_excel(newpath)
    elif (newpath.split("/")[-1].split(".")[-1] == "csv"):
        if option=='default':
          #df=pd.read_csv(newpath).reset_index()
          df = clevercsv.read_dataframe(newpath).reset_index()
          #df=df.T.reset_index().T
        elif option == 1:
          with open(newpath, "r", newline="") as fp:
            # you can use verbose=True to see what CleverCSV does
            dialect = clevercsv.Sniffer().sniff(fp.read(), verbose=False)
            fp.seek(0)
            reader = clevercsv.reader(fp, dialect)
            rows = list(reader)
          df=pd.DataFrame(rows)
          df.columns=["a"+str(i) for i in df.columns]
        elif option==2:
          df = clevercsv.read_dataframe(newpath)
        elif option==3:
          rows = clevercsv.read_table(newpath)
          df=pd.DataFrame(rows)  
    else:
        raise ValueError('Format Not Found')
    
    if redundancy:
      df=df.dropna(thresh=len(df)-round(0.99*round(len(df))), axis=1)
    display(df)
    newpath=input_path('Input folder to save:')
    randomo=random.randint(100000, 1000000)
    try:
      pyspssio.write_sav(newpath+f"/data_{randomo}.sav", df)
    except:
      pyreadstat.write_sav(df, newpath+f"/data_{randomo}.sav")
    print(f'Saved as data_{randomo}')
    print('Wait some minutes, the Google Drive is updating')
    return df

def fix_csv(option='default'): #try to fix csv
    #import drive
    drive.mount("/content/drive")
    newpath=input_path('Input path:')
    if (newpath.split("/")[-1].split(".")[-1] == "xlsx") or (newpath.split("/")[-1].split(".")[-1] == "xls"):
        df=pd.read_excel(newpath)
    elif (newpath.split("/")[-1].split(".")[-1] == "csv"):
        if option=='default':
          with open(newpath, "r", newline="") as fp:
            # you can use verbose=True to see what CleverCSV does
            dialect = clevercsv.Sniffer().sniff(fp.read(), verbose=False)
            fp.seek(0)
            reader = clevercsv.reader(fp, dialect)
            rows = list(reader)
          df=pd.DataFrame(rows)
        elif option==1:
          df = clevercsv.read_dataframe(newpath)
        elif option==2:
          rows = clevercsv.read_table(newpath)
          df=pd.DataFrame(rows)  
    else:
        raise ValueError('Format Not Found')
    
    df=df.dropna(thresh=len(df)-round(0.99*round(len(df))), axis=1)
    display(df)
    newpath=input_path('Input folder to save:')
    randomo=random.randint(100000, 1000000)
    df.to_csv(newpath+f"/data_{randomo}.csv",index=False)
    print(f'Saved as data_{randomo}')
    print('Wait some minutes, the Google Drive is updating')
